/* Simple RIPE to I: line converter
 *
 * Copyright 2001 - 2002 pht <pht@system.cz>
 * Please send comments, patches and ideas.
 *
 * Licensed under GPL. If you don't know or have it, you can request it at:
 * Free Software Foundation, Inc.  675 Mass Ave, Cambridge, MA 02139, USA
 *
 * EFnet support written by W. Campbell <wcampbel@botbay.net>
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>

#ifdef __CYGWIN__
#include <getopt.h>
#endif

/* This struct is used in config.h */
struct i_info {                 /* fill this to tune I line generation */
        char *country;          /* RIPE country */
        int big_i;              /* 0 = i: (restricted), 1 = I: */
        char *prefix;           /* I: line prefix, H6 and +hemp */
        char *y_class;          /* connection class (char to support H7) */
};

#include "config.h"

#define EL "\033[K"
#define HASH_SIZE 256 * 16

struct ripe_data {
	unsigned long net;
	int mask;
	char *country;
	char *prefix;
	char *y_class;
	int times;
	int count;
	int big_i;
	char descr[500];
	char netname[200];
} current;

struct ripe {
	struct ripe *next, *merge;
	struct ripe_data data;
} *hash[HASH_SIZE];

unsigned long in_lines = 0, out_lines = 0, nets_found = 0, nets_dup = 0,
	      nets_merge = 0, i_lines = 0, nets_ok = 0;
char *infile = NULL, *outfile = NULL;
FILE *inf, *outf;

void reset_current()
{
	memset(&current, 0, sizeof(current));
}

void reset_hash()
{
	memset(&hash, 0, sizeof(hash));
}

long hash_func(long x)
{
	x >>= 16;
	x &= 0x00000fffUL;
	return x;
}

unsigned long mask_addr(unsigned long net, int mask)
{
	net >>= 32 - mask;
	net <<= 32 - mask;
	return net;
}

struct ripe *new_ripe()
{
	struct ripe *p;

	p = malloc(sizeof(struct ripe));

	if (!p) {
		fprintf(stderr, EL "Aborting due to insufficient memory.\n");
		exit(1);
	}

	p->next = p->merge = NULL;

	return p;
}

int equ_class(struct ripe_data *p, struct ripe_data *q)
{
        return (p->y_class == q->y_class) && (p->big_i == q->big_i);
}

int search_dups_single_hash(unsigned long h)
{
        unsigned long pnet, cnet;
        int mask, p_smaller;
        struct ripe *p, *o;

        for (p = hash[h], o = NULL; p; p = p->next) {
                p_smaller = p->data.mask < current.mask;
                mask = p_smaller ? p->data.mask : current.mask;
                pnet = mask_addr(p->data.net, mask);
                cnet = mask_addr(current.net, mask);
                if ((pnet == cnet) && equ_class(&current, &(p->data))) {
                        nets_dup++;
                        fprintf(stderr, EL "Dup found: %s %c %s\r",
                                p->data.netname, p_smaller ? '>' : '<',
                                current.netname);
                        if (p_smaller)
                                return 1; /* order current to be killed */
                        if (o)
                                o->next = p->next; /* kill p */
                        else
                                hash[h] = p->next;
                } else
                        o = p;
        }
        return 0;
}

int can_merge(unsigned long net1, unsigned long net2, int mask)
{
        unsigned long tmp;

        tmp = net1 ^ net2;
        tmp >>= 33 - mask;

        return tmp == 0;
}

void merge_nets(struct ripe *p, struct ripe *q)
{
        struct ripe *t;
        
        nets_merge++;

        for (t = p; t->merge; t = t->merge);
        t->merge = q;

        p->data.mask--;
        p->data.net = mask_addr(p->data.net, p->data.mask);
        fprintf(stderr, EL "Merged: %s + %s\r", p->data.netname,
                q->data.netname);
}

int search_merges_single_hash(unsigned long h, struct ripe *p)
{
        struct ripe *q, *o;
        
        for (q = hash[h], o = NULL; q; o = q, q = q->next) {
                if ((p != q) && (q->data.mask == p->data.mask)) {
                        if (can_merge(p->data.net, q->data.net,
                            p->data.mask) &&
                            equ_class(&(p->data), &(q->data))) {
                                if (o)
                                        o->next = q->next;
                                else
                                        hash[h] = q->next;
                                merge_nets(p, q);
                                return 1;
                        }
                }
        }
        return 0;
}

int search(int what, struct ripe *p) /* 0 = dups, 1 = merges */
{
	unsigned long h;
	int r;
	
	for (h = 0; h < HASH_SIZE; h++) {
		if (what)
			r = search_merges_single_hash(h, p);
		else
			r = search_dups_single_hash(h);
		if (r)
			return r;
	}
	return 0;
}

void add_to_hash()
{
        unsigned long h;
        struct ripe *p;
                
        nets_found++;
        if (current.y_class && current.mask && current.net
            && *(current.netname) && current.prefix && current.country)
                for (current.count = current.times; current.count;
                                current.count--) {
                        nets_ok++;
#ifdef SEARCH_FOR_DUPS
                        if (!search(0, NULL)) { /* kill inclusive net(s) */
#endif
                                h = hash_func(current.net);
                                p = new_ripe();
                                p->next = hash[h];
                                hash[h] = p;
                                memcpy(&(p->data), &current,
                                                sizeof(struct ripe_data));
#ifdef SEARCH_FOR_MERGES
                                while (search(1, p)); /* look if it can't be
                                                    merged with other net(s) */
#endif
#ifdef SEARCH_FOR_DUPS
                        }
#endif
                        current.net += 1L << (32 - current.mask);
                }
        reset_current();
}

void print_netinfo(struct ripe_data *data, int cont)
{
	char *comment;

#ifdef IRCD_CS 
	comment = "//";
#else
	comment = "#";
#endif
        if (cont < 2) {
                fprintf(outf,"%s %s @RIPE (%s)", comment,data->netname, 
			*(data->descr) ? data->descr : "*unknown*");
        } else
                fprintf(outf,"%s (%d more sub-blocks) ... %s @RIPE",comment, 
		cont, data->netname);
        
        if (data->times > 1) {
                fprintf(outf," [sub-block %d/%d]", data->count, data->times);
        }

        out_lines++;
        fprintf(outf,"\n");
}

void commit_iline(struct in_addr i, struct ripe *p)
{
	char mbuf[10];
#ifdef WORKAROUND_32
	if (p->data.mask == 32)
		mbuf[0] = 0;
	else
#endif
		sprintf(mbuf,"/%d",p->data.mask);
#ifdef IRCD_IRCNET
       	fprintf(outf,"%c:%s%s::%s::%s\n", p->data.big_i ?  'I' : 'i', 
			inet_ntoa(i), mbuf,p->data.prefix,p->data.y_class);
	out_lines++;

#endif
#ifdef IRCD_HYB6
	fprintf(outf,"%c:%s%s::%sx::%s\n", p->data.big_i ? 'I' : 'i',
				inet_ntoa(i), mbuf,p->data.prefix,
				p->data.y_class);
	out_lines++;
#endif
#ifdef IRCD_HYB7
	fprintf(outf, "auth {\n");
	fprintf(outf, "\tuser = \"*@%s%s\";\n", inet_ntoa(i),
	       mbuf);
	fprintf(outf, "\tclass = \"%s\";\n", p->data.y_class);
	if (p->data.big_i == 0) {
		fprintf(outf, "\trestricted = yes;\n");
		out_lines++;
	}
	fprintf(outf, "};\n\n");
	out_lines += 5;
#endif
#ifdef IRCD_CS
	fprintf(outf, "allow {\n");
	fprintf(outf, "\tipaddr %s%s;\n", inet_ntoa(i),mbuf);
	fprintf(outf, "\tclass %s;\n", p->data.y_class);
	fprintf(outf, "};\n\n");
	out_lines += 5
#endif

}
void commit()
{
        struct in_addr i;
        struct ripe *p, *q, *o;
        long h, next_print = 0;
        int cont;

        for (h = 0; h < HASH_SIZE; h++)
                for (p = hash[h]; p; p = p->next) {
                        i.s_addr = htonl(p->data.net);
                        if (p->merge) {
                                out_lines++;
                                fprintf(outf,"%s merged block of:\n",
#ifdef IRCD_CS
						"//"
#else
						"#"
#endif
						);
                                cont = 0;
                                for (o = NULL, q = p; q; o = q, q = q->merge) {
                                        if (o && o->data.count == q->data.count 
- 1) {
                                                cont++;
                                        } else {
                                                print_netinfo(&(q->data), cont);
                                                cont = 0;
                                        }
                                }
                                if (cont)
                                        print_netinfo(&(o->data), cont);
                        } else {
                                print_netinfo(&(p->data), 0);
                        }
                        i_lines++;
			commit_iline(i,p);
                        if (out_lines > next_print) {
                                fprintf(stderr, EL "Writing: %luw / %luI\r",
                                        out_lines, i_lines);
                                fflush(stderr);
                                next_print += 1000;
                        }
                }
}

void do_inetnum(char *p)
{
        char *q;
        struct in_addr i;
        unsigned long start, end, tmp;

        if (current.net)
                return;

        q = strchr(p, ' ');
        if (!q)
                return;
        *q = 0;
        do q++; while (*q == ' ' || *q == '\t' || *q == '-');

        if (!inet_aton(p, &i))
                return;
        start = ntohl(i.s_addr);
    
        if (!inet_aton(q, &i))
                return;
        end = ntohl(i.s_addr);

        if (end < start) {
                fprintf(stderr, EL "Broken inetnum at line %ld\n", in_lines);
                return;
        }
        tmp = end - start;
	
#ifdef RIPE_NL_WORKAROUND
	/* skip 0.0.0.0 - 255.255.255.25 entry */
	if (!start && end == 0xffffffffUL)
	{
		return;
	}
		
#endif
#ifdef OFF_BY_ONE_WORKAROUND
	if (tmp > 1) 
	{
		int flag = 0;
		if (start & 1)
		{
			flag = 1;
			start--;
			tmp++;
		}
		if (!(end & 1))
		{
			flag += 2;
			end++;
			tmp++;
		}
		if (flag)
		{
		    fprintf(stderr, EL "Fixing off by one(%d) at line %ld \n [%s - %s] \n", flag,in_lines,p,q);
		}
	}
#endif

        current.mask = 32;
        while (tmp & 1) {
                tmp >>= 1;
                current.mask--;
        }

        current.times = 1;
        if (tmp)
                current.times += tmp;

        current.net = mask_addr(start, current.mask);
        while (current.net != start) {
                current.mask++;
                current.times <<= 1;
                current.net = mask_addr(start, current.mask);
        }
}

void do_netname(char *p)
{
        if (!*current.netname) {
                strncpy(current.netname, p, sizeof(current.netname));
                current.netname[sizeof(current.netname) - 1] = 0;
        }
}


void do_descr(char *p)
{
	int l, s;

	l = strlen(current.descr);
	s = sizeof(current.descr);

	if (l == s - 1)
		return;
	
	if (*current.descr) {
		strncat(current.descr, " / ", s - l - 1);
		l += 3;
	}
	strncat(current.descr, p, s - l - 1);
	current.descr[sizeof(current.descr) - 1] = 0;
}

void do_country(char *p)
{
	struct i_info *i;

	for (i = i_info; i->country; i++) {
		if (!strcmp(i->country, p)) {
			current.country = i->country;
			current.big_i = i->big_i;
			current.y_class = i->y_class;
			current.prefix = i->prefix;
			return;
		}
	}
}

void parse_line(char *line)
{
        char *p;

        if (*line == '#')
                return;

        p = strchr(line, ':');
        if (!p) {
                p = line;
                while (*p == ' ' || *p == '\t')
                        p++;
                if (*p)
                        fprintf(stderr, EL "No `:' on line %lu: %s\r",
                                in_lines, line);
                else
                        add_to_hash();
                return;
        }
        *p = 0;
        do p++; while (*p == ' ' || *p == '\t');
        if (!*p) {
                fprintf(stderr, EL "Null entry on line %lu: %s\r", in_lines,
                        line);
                return;
        }

        if (!strcmp(line, "inetnum"))
                do_inetnum(p);
        if (!strcmp(line, "netname"))
                do_netname(p);
        if (!strcmp(line, "descr"))
                do_descr(p);
        if (!strcmp(line, "country"))
                do_country(p);
}

void parse_file()
{
	char buf[4096], *p;

	while (fgets(buf, sizeof(buf), inf)) {
		in_lines++;
		p = strchr(buf, '\n');
		if (p)
			*p = 0;
		else
			fprintf(stderr, EL "Too long line: %lu\n", in_lines);
		parse_line(buf);
		if (!(in_lines % 10000)) {
			fprintf(stderr, EL "Parsing: %lur / %lun\r", in_lines,
				nets_found);
			fflush(stderr);
		}
	}
	parse_line("");
}

void cmd_line(int argc, char *argv[])
{
	char c;
	extern char *optarg;
	/* extern int optind; */

	/* This section sets the default input and output files.
	 * The old code used stdin for the database and stdout for
	 * the output file.  It will still work that way, but allowing
	 * the user to specify a filename for both is easier
	 */
	inf = stdin;
	outf = stdout;

	while( (c=getopt(argc, argv, "d:i:o:h?")) != -1)
	{
	    switch (c) {
		case 'd':
		case 'i':
			inf = fopen(optarg, "r");
			if (inf == NULL) {
				fprintf(stderr, EL "Unable to open database: "
					"%s\n", optarg);
				exit(1);
			}
			break;
		case 'o':
			outf = fopen(optarg, "w");
			if (outf == NULL) {
				fprintf(stderr, EL "Unable to open output "
					"file: %s\n", optarg);
				exit(2);
			}
			break;
		case 'h':
		case '?':
			fprintf(stderr, EL "%s [-d database] [-o output] "
				"[-h|-?]\n -d - Specify a database to use\n"
				" -o - Spefify an output file to use\n",
				argv[0]);
			exit(0);
			/* NOT REACHED */
			break;
		default:
			fprintf(stderr, EL "Invalid Option: -%c\n", c);
			break;
	    }
	}
}

int main(int argc, char *argv[])
{
	cmd_line(argc, argv);
	reset_current();
	reset_hash();
	parse_file();
	commit();
	fprintf(stderr, EL "\nFinal statistics:\n");
	fprintf(stderr, "Lines read                   : %lu\n", in_lines);
	fprintf(stderr, "Total networks found         : %lu\n", nets_found);
	fprintf(stderr, "Acceptable networks found    : %lu\n", nets_ok);
	fprintf(stderr, "Duplicate/inclusive networks : %lu\n", nets_dup);
	fprintf(stderr, "Merged networks              : %lu\n", nets_merge);
	fprintf(stderr, "I-lines written              : %lu\n", i_lines);
	fprintf(stderr, "Lines written                : %lu\n", out_lines);
	return 0;
}
